@@ -1,21 +1,19 @@
 # Copyright (c) 2017 Ultimaker B.V.
 # Cura is released under the terms of the LGPLv3 or higher.
 
-from PyQt5.QtCore import pyqtSignal, pyqtProperty, QObject, QVariant #For communicating data and events to Qt.
+from PyQt5.QtCore import pyqtSignal, pyqtProperty, QObject, QVariant  # For communicating data and events to Qt.
 from UM.FlameProfiler import pyqtSlot
 
-from UM.Application import Application #To get the global container stack to find the current machine.
+from UM.Application import Application  # To get the global container stack to find the current machine.
 from UM.Logger import Logger
-from UM.Decorators import deprecated
 from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator
 from UM.Scene.SceneNode import SceneNode
 from UM.Scene.Selection import Selection
 from UM.Scene.Iterator.BreadthFirstIterator import BreadthFirstIterator
-from UM.Settings.ContainerRegistry import ContainerRegistry #Finding containers by ID.
-from UM.Settings.InstanceContainer import InstanceContainer
+from UM.Settings.ContainerRegistry import ContainerRegistry  # Finding containers by ID.
 from UM.Settings.SettingFunction import SettingFunction
+from UM.Settings.SettingInstance import SettingInstance
 from UM.Settings.ContainerStack import ContainerStack
-from UM.Settings.Interfaces import DefinitionContainerInterface
 from UM.Settings.PropertyEvaluationContext import PropertyEvaluationContext
 from typing import Optional, List, TYPE_CHECKING, Union
 
@@ -28,28 +26,27 @@ if TYPE_CHECKING:
 #
 #   This keeps a list of extruder stacks for each machine.
 class ExtruderManager(QObject):
-    ##  Signal to notify other components when the list of extruders for a machine definition changes.
-    extrudersChanged = pyqtSignal(QVariant)
-
-    ## Signal to notify other components when the global container stack is switched to a definition
-    #  that has different extruders than the previous global container stack
-    globalContainerStackDefinitionChanged = pyqtSignal()
-
-    ##  Notify when the user switches the currently active extruder.
-    activeExtruderChanged = pyqtSignal()
 
     ##  Registers listeners and such to listen to changes to the extruders.
     def __init__(self, parent = None):
         super().__init__(parent)
-        self._extruder_trains = { } #Per machine, a dictionary of extruder container stack IDs. Only for separately defined extruders.
-        self._active_extruder_index = -1 # Indicates the index of the active extruder stack. -1 means no active extruder stack
+
+        self._application = Application.getInstance()
+
+        self._extruder_trains = {}  # Per machine, a dictionary of extruder container stack IDs. Only for separately defined extruders.
+        self._active_extruder_index = -1  # Indicates the index of the active extruder stack. -1 means no active extruder stack
         self._selected_object_extruders = []
-        Application.getInstance().globalContainerStackChanged.connect(self.__globalContainerStackChanged)
-        self._global_container_stack_definition_id = None
         self._addCurrentMachineExtruders()
 
+        #Application.getInstance().globalContainerStackChanged.connect(self._globalContainerStackChanged)
         Selection.selectionChanged.connect(self.resetSelectedObjectExtruders)
 
+    ##  Signal to notify other components when the list of extruders for a machine definition changes.
+    extrudersChanged = pyqtSignal(QVariant)
+
+    ##  Notify when the user switches the currently active extruder.
+    activeExtruderChanged = pyqtSignal()
+
     ##  Gets the unique identifier of the currently active extruder stack.
     #
     #   The currently active extruder stack is the stack that is currently being
@@ -59,10 +56,10 @@ class ExtruderManager(QObject):
     @pyqtProperty(str, notify = activeExtruderChanged)
     def activeExtruderStackId(self) -> Optional[str]:
         if not Application.getInstance().getGlobalContainerStack():
-            return None # No active machine, so no active extruder.
+            return None  # No active machine, so no active extruder.
         try:
             return self._extruder_trains[Application.getInstance().getGlobalContainerStack().getId()][str(self._active_extruder_index)].getId()
-        except KeyError: # Extruder index could be -1 if the global tab is selected, or the entry doesn't exist if the machine definition is wrong.
+        except KeyError:  # Extruder index could be -1 if the global tab is selected, or the entry doesn't exist if the machine definition is wrong.
             return None
 
     ##  Return extruder count according to extruder trains.
@@ -76,23 +73,23 @@ class ExtruderManager(QObject):
             return 0
 
     ##  Gets a dict with the extruder stack ids with the extruder number as the key.
-    #   The key "-1" indicates the global stack id.
-    #
     @pyqtProperty("QVariantMap", notify = extrudersChanged)
     def extruderIds(self):
         extruder_stack_ids = {}
+
         global_stack_id = Application.getInstance().getGlobalContainerStack().getId()
-        extruder_stack_ids["-1"] = global_stack_id
+
         if global_stack_id in self._extruder_trains:
             for position in self._extruder_trains[global_stack_id]:
                 extruder_stack_ids[position] = self._extruder_trains[global_stack_id][position].getId()
+
         return extruder_stack_ids
 
     @pyqtSlot(str, result = str)
-    def getQualityChangesIdByExtruderStackId(self, id: str) -> str:
+    def getQualityChangesIdByExtruderStackId(self, extruder_stack_id: str) -> str:
         for position in self._extruder_trains[Application.getInstance().getGlobalContainerStack().getId()]:
             extruder = self._extruder_trains[Application.getInstance().getGlobalContainerStack().getId()][position]
-            if extruder.getId() == id:
+            if extruder.getId() == extruder_stack_id:
                 return extruder.qualityChanges.getId()
 
     ##  The instance of the singleton pattern.
@@ -100,6 +97,10 @@ class ExtruderManager(QObject):
     #   It's None if the extruder manager hasn't been created yet.
     __instance = None
 
+    @staticmethod
+    def createExtruderManager():
+        return ExtruderManager().getInstance()
+
     ##  Gets an instance of the extruder manager, or creates one if no instance
     #   exists yet.
     #
@@ -178,6 +179,7 @@ class ExtruderManager(QObject):
         self._selected_object_extruders = []
         self.selectedObjectExtrudersChanged.emit()
 
+    @pyqtSlot(result = QObject)
     def getActiveExtruderStack(self) -> Optional["ExtruderStack"]:
         global_container_stack = Application.getInstance().getGlobalContainerStack()
 
@@ -185,6 +187,7 @@ class ExtruderManager(QObject):
             if global_container_stack.getId() in self._extruder_trains:
                 if str(self._active_extruder_index) in self._extruder_trains[global_container_stack.getId()]:
                     return self._extruder_trains[global_container_stack.getId()][str(self._active_extruder_index)]
+
         return None
 
     ##  Get an extruder stack by index
@@ -203,40 +206,6 @@ class ExtruderManager(QObject):
             result.append(self.getExtruderStack(i))
         return result
 
-    ##  Adds all extruders of a specific machine definition to the extruder
-    #   manager.
-    #
-    #   \param machine_definition   The machine definition to add the extruders for.
-    #   \param machine_id           The machine_id to add the extruders for.
-    @deprecated("Use CuraStackBuilder", "2.6")
-    def addMachineExtruders(self, machine_definition: DefinitionContainerInterface, machine_id: str) -> None:
-        changed = False
-        machine_definition_id = machine_definition.getId()
-        if machine_id not in self._extruder_trains:
-            self._extruder_trains[machine_id] = { }
-            changed = True
-        container_registry = ContainerRegistry.getInstance()
-        if container_registry:
-            # Add the extruder trains that don't exist yet.
-            for extruder_definition in container_registry.findDefinitionContainers(machine = machine_definition_id):
-                position = extruder_definition.getMetaDataEntry("position", None)
-                if not position:
-                    Logger.log("w", "Extruder definition %s specifies no position metadata entry.", extruder_definition.getId())
-                if not container_registry.findContainerStacks(machine = machine_id, position = position): # Doesn't exist yet.
-                    self.createExtruderTrain(extruder_definition, machine_definition, position, machine_id)
-                    changed = True
-
-            # Gets the extruder trains that we just created as well as any that still existed.
-            extruder_trains = container_registry.findContainerStacks(type = "extruder_train", machine = machine_id)
-            for extruder_train in extruder_trains:
-                self._extruder_trains[machine_id][extruder_train.getMetaDataEntry("position")] = extruder_train
-
-                # regardless of what the next stack is, we have to set it again, because of signal routing.
-                extruder_train.setNextStack(Application.getInstance().getGlobalContainerStack())
-                changed = True
-        if changed:
-            self.extrudersChanged.emit(machine_id)
-
     def registerExtruder(self, extruder_train, machine_id):
         changed = False
 
@@ -256,138 +225,6 @@ class ExtruderManager(QObject):
         if changed:
             self.extrudersChanged.emit(machine_id)
 
-    ##  Creates a container stack for an extruder train.
-    #
-    #   The container stack has an extruder definition at the bottom, which is
-    #   linked to a machine definition. Then it has a variant profile, a material
-    #   profile, a quality profile and a user profile, in that order.
-    #
-    #   The resulting container stack is added to the registry.
-    #
-    #   \param extruder_definition  The extruder to create the extruder train for.
-    #   \param machine_definition   The machine that the extruder train belongs to.
-    #   \param position             The position of this extruder train in the extruder slots of the machine.
-    #   \param machine_id           The id of the "global" stack this extruder is linked to.
-    @deprecated("Use CuraStackBuilder::createExtruderStack", "2.6")
-    def createExtruderTrain(self, extruder_definition: DefinitionContainerInterface, machine_definition: DefinitionContainerInterface,
-                            position, machine_id: str) -> None:
-        # Cache some things.
-        container_registry = ContainerRegistry.getInstance()
-        machine_definition_id = Application.getInstance().getMachineManager().getQualityDefinitionId(machine_definition)
-
-        # Create a container stack for this extruder.
-        extruder_stack_id = container_registry.uniqueName(extruder_definition.getId())
-        container_stack = ContainerStack(extruder_stack_id)
-        container_stack.setName(extruder_definition.getName())  # Take over the display name to display the stack with.
-        container_stack.addMetaDataEntry("type", "extruder_train")
-        container_stack.addMetaDataEntry("machine", machine_id)
-        container_stack.addMetaDataEntry("position", position)
-        container_stack.addContainer(extruder_definition)
-
-        # Find the variant to use for this extruder.
-        variant = container_registry.findInstanceContainers(id = "empty_variant")[0]
-        if machine_definition.getMetaDataEntry("has_variants"):
-            # First add any variant. Later, overwrite with preference if the preference is valid.
-            variants = container_registry.findInstanceContainers(definition = machine_definition_id, type = "variant")
-            if len(variants) >= 1:
-                variant = variants[0]
-            preferred_variant_id = machine_definition.getMetaDataEntry("preferred_variant")
-            if preferred_variant_id:
-                preferred_variants = container_registry.findInstanceContainers(id = preferred_variant_id, definition = machine_definition_id, type = "variant")
-                if len(preferred_variants) >= 1:
-                    variant = preferred_variants[0]
-                else:
-                    Logger.log("w", "The preferred variant \"%s\" of machine %s doesn't exist or is not a variant profile.", preferred_variant_id, machine_id)
-                    # And leave it at the default variant.
-        container_stack.addContainer(variant)
-
-        # Find a material to use for this variant.
-        material = container_registry.findInstanceContainers(id = "empty_material")[0]
-        if machine_definition.getMetaDataEntry("has_materials"):
-            # First add any material. Later, overwrite with preference if the preference is valid.
-            machine_has_variant_materials = machine_definition.getMetaDataEntry("has_variant_materials", default = False)
-            if machine_has_variant_materials or machine_has_variant_materials == "True":
-                materials = container_registry.findInstanceContainers(type = "material", definition = machine_definition_id, variant = variant.getId())
-            else:
-                materials = container_registry.findInstanceContainers(type = "material", definition = machine_definition_id)
-            if len(materials) >= 1:
-                material = materials[0]
-            preferred_material_id = machine_definition.getMetaDataEntry("preferred_material")
-            if preferred_material_id:
-                global_stack = ContainerRegistry.getInstance().findContainerStacks(id = machine_id)
-                if global_stack:
-                    approximate_material_diameter = str(round(global_stack[0].getProperty("material_diameter", "value")))
-                else:
-                    approximate_material_diameter = str(round(machine_definition.getProperty("material_diameter", "value")))
-
-                search_criteria = { "type": "material",  "id": preferred_material_id, "approximate_diameter": approximate_material_diameter}
-                if machine_definition.getMetaDataEntry("has_machine_materials"):
-                    search_criteria["definition"] = machine_definition_id
-
-                    if machine_definition.getMetaDataEntry("has_variants") and variant:
-                        search_criteria["variant"] = variant.id
-                else:
-                    search_criteria["definition"] = "fdmprinter"
-
-                preferred_materials = container_registry.findInstanceContainers(**search_criteria)
-                if len(preferred_materials) >= 1:
-                    # In some cases we get multiple materials. In that case, prefer materials that are marked as read only.
-                    read_only_preferred_materials = [preferred_material for preferred_material in preferred_materials if preferred_material.isReadOnly()]
-                    if len(read_only_preferred_materials) >= 1:
-                        material = read_only_preferred_materials[0]
-                    else:
-                        material = preferred_materials[0]
-                else:
-                    Logger.log("w", "The preferred material \"%s\" of machine %s doesn't exist or is not a material profile.", preferred_material_id, machine_id)
-                    # And leave it at the default material.
-        container_stack.addContainer(material)
-
-        # Find a quality to use for this extruder.
-        quality = container_registry.getEmptyInstanceContainer()
-
-        search_criteria = { "type": "quality" }
-        if machine_definition.getMetaDataEntry("has_machine_quality"):
-            search_criteria["definition"] = machine_definition_id
-            if machine_definition.getMetaDataEntry("has_materials") and material:
-                search_criteria["material"] = material.id
-        else:
-            search_criteria["definition"] = "fdmprinter"
-
-        preferred_quality = machine_definition.getMetaDataEntry("preferred_quality")
-        if preferred_quality:
-            search_criteria["id"] = preferred_quality
-
-        containers = ContainerRegistry.getInstance().findInstanceContainers(**search_criteria)
-        if not containers and preferred_quality:
-            Logger.log("w", "The preferred quality \"%s\" of machine %s doesn't exist or is not a quality profile.", preferred_quality, machine_id)
-            search_criteria.pop("id", None)
-            containers = ContainerRegistry.getInstance().findInstanceContainers(**search_criteria)
-        if containers:
-            quality = containers[0]
-
-        container_stack.addContainer(quality)
-
-        empty_quality_changes = container_registry.findInstanceContainers(id = "empty_quality_changes")[0]
-        container_stack.addContainer(empty_quality_changes)
-
-        user_profile = container_registry.findInstanceContainers(type = "user", extruder = extruder_stack_id)
-        if user_profile: # There was already a user profile, loaded from settings.
-            user_profile = user_profile[0]
-        else:
-            user_profile = InstanceContainer(extruder_stack_id + "_current_settings")  # Add an empty user profile.
-            user_profile.addMetaDataEntry("type", "user")
-            user_profile.addMetaDataEntry("extruder", extruder_stack_id)
-            from cura.CuraApplication import CuraApplication
-            user_profile.addMetaDataEntry("setting_version", CuraApplication.SettingVersion)
-            user_profile.setDefinition(machine_definition)
-            container_registry.addContainer(user_profile)
-        container_stack.addContainer(user_profile)
-
-        # regardless of what the next stack is, we have to set it again, because of signal routing.
-        container_stack.setNextStack(Application.getInstance().getGlobalContainerStack())
-
-        container_registry.addContainer(container_stack)
-
     def getAllExtruderValues(self, setting_key):
         return self.getAllExtruderSettings(setting_key, "value")
 
@@ -396,18 +233,21 @@ class ExtruderManager(QObject):
     #   \param setting_key  \type{str} The setting to get the property of.
     #   \param property  \type{str} The property to get.
     #   \return \type{List} the list of results
-    def getAllExtruderSettings(self, setting_key, property):
-        global_container_stack = Application.getInstance().getGlobalContainerStack()
-        if global_container_stack.getProperty("machine_extruder_count", "value") <= 1:
-            return [global_container_stack.getProperty(setting_key, property)]
-
+    def getAllExtruderSettings(self, setting_key: str, prop: str):
         result = []
         for index in self.extruderIds:
             extruder_stack_id = self.extruderIds[str(index)]
-            stack = ContainerRegistry.getInstance().findContainerStacks(id = extruder_stack_id)[0]
-            result.append(stack.getProperty(setting_key, property))
+            extruder_stack = ContainerRegistry.getInstance().findContainerStacks(id = extruder_stack_id)[0]
+            result.append(extruder_stack.getProperty(setting_key, prop))
         return result
 
+    def extruderValueWithDefault(self, value):
+        machine_manager = self._application.getMachineManager()
+        if value == "-1":
+            return machine_manager.defaultExtruderPosition
+        else:
+            return value
+
     ##  Gets the extruder stacks that are actually being used at the moment.
     #
     #   An extruder stack is being used if it is the extruder to print any mesh
@@ -419,23 +259,28 @@ class ExtruderManager(QObject):
     #
     #   \return A list of extruder stacks.
     def getUsedExtruderStacks(self) -> List["ContainerStack"]:
-        global_stack = Application.getInstance().getGlobalContainerStack()
+        global_stack = self._application.getGlobalContainerStack()
         container_registry = ContainerRegistry.getInstance()
 
-        if global_stack.getProperty("machine_extruder_count", "value") <= 1: #For single extrusion.
-            return [global_stack]
-
         used_extruder_stack_ids = set()
 
-        #Get the extruders of all meshes in the scene.
+        # Get the extruders of all meshes in the scene
         support_enabled = False
         support_bottom_enabled = False
         support_roof_enabled = False
+
         scene_root = Application.getInstance().getController().getScene().getRoot()
-        meshes = [node for node in DepthFirstIterator(scene_root) if type(node) is SceneNode and node.isSelectable()] #Only use the nodes that will be printed.
+
+        # If no extruders are registered in the extruder manager yet, return an empty array
+        if len(self.extruderIds) == 0:
+            return []
+
+        # Get the extruders of all printable meshes in the scene
+        meshes = [node for node in DepthFirstIterator(scene_root) if isinstance(node, SceneNode) and node.isSelectable()]
         for mesh in meshes:
             extruder_stack_id = mesh.callDecoration("getActiveExtruder")
-            if not extruder_stack_id: #No per-object settings for this node.
+            if not extruder_stack_id:
+                # No per-object settings for this node
                 extruder_stack_id = self.extruderIds["0"]
             used_extruder_stack_ids.add(extruder_stack_id)
 
@@ -464,16 +309,20 @@ class ExtruderManager(QObject):
 
         # Check support extruders
         if support_enabled:
-            used_extruder_stack_ids.add(self.extruderIds[str(global_stack.getProperty("support_infill_extruder_nr", "value"))])
-            used_extruder_stack_ids.add(self.extruderIds[str(global_stack.getProperty("support_extruder_nr_layer_0", "value"))])
+            used_extruder_stack_ids.add(self.extruderIds[self.extruderValueWithDefault(str(global_stack.getProperty("support_infill_extruder_nr", "value")))])
+            used_extruder_stack_ids.add(self.extruderIds[self.extruderValueWithDefault(str(global_stack.getProperty("support_extruder_nr_layer_0", "value")))])
             if support_bottom_enabled:
-                used_extruder_stack_ids.add(self.extruderIds[str(global_stack.getProperty("support_bottom_extruder_nr", "value"))])
+                used_extruder_stack_ids.add(self.extruderIds[self.extruderValueWithDefault(str(global_stack.getProperty("support_bottom_extruder_nr", "value")))])
             if support_roof_enabled:
-                used_extruder_stack_ids.add(self.extruderIds[str(global_stack.getProperty("support_roof_extruder_nr", "value"))])
+                used_extruder_stack_ids.add(self.extruderIds[self.extruderValueWithDefault(str(global_stack.getProperty("support_roof_extruder_nr", "value")))])
 
-        #The platform adhesion extruder. Not used if using none.
+        # The platform adhesion extruder. Not used if using none.
         if global_stack.getProperty("adhesion_type", "value") != "none":
-            used_extruder_stack_ids.add(self.extruderIds[str(global_stack.getProperty("adhesion_extruder_nr", "value"))])
+            extruder_nr = str(global_stack.getProperty("adhesion_extruder_nr", "value"))
+            if extruder_nr == "-1":
+                extruder_nr = Application.getInstance().getMachineManager().defaultExtruderPosition
+            used_extruder_stack_ids.add(self.extruderIds[extruder_nr])
+
         try:
             return [container_registry.findContainerStacks(id = stack_id)[0] for stack_id in used_extruder_stack_ids]
         except IndexError:  # One or more of the extruders was not found.
@@ -516,44 +365,121 @@ class ExtruderManager(QObject):
     #   \return \type{List[ContainerStack]} a list of
     def getActiveExtruderStacks(self) -> List["ExtruderStack"]:
         global_stack = Application.getInstance().getGlobalContainerStack()
+        if not global_stack:
+            return None
 
         result = []
-        machine_extruder_count = global_stack.getProperty("machine_extruder_count", "value")
-
-        # In case the printer is using one extruder, shouldn't exist active extruder stacks
-        if machine_extruder_count == 1:
-            return result
-
-        if global_stack and global_stack.getId() in self._extruder_trains:
+        if global_stack.getId() in self._extruder_trains:
             for extruder in sorted(self._extruder_trains[global_stack.getId()]):
                 result.append(self._extruder_trains[global_stack.getId()][extruder])
 
+        machine_extruder_count = global_stack.getProperty("machine_extruder_count", "value")
+
         return result[:machine_extruder_count]
 
-    def __globalContainerStackChanged(self) -> None:
-        global_container_stack = Application.getInstance().getGlobalContainerStack()
-        if global_container_stack and global_container_stack.getBottom() and global_container_stack.getBottom().getId() != self._global_container_stack_definition_id:
-            self._global_container_stack_definition_id = global_container_stack.getBottom().getId()
-            self.globalContainerStackDefinitionChanged.emit()
-
-        # If the global container changed, the number of extruders could be changed and so the active_extruder_index is updated
-        extruder_count = global_container_stack.getProperty("machine_extruder_count", "value")
-        if extruder_count > 1:
-            if self._active_extruder_index == -1:
-                self.setActiveExtruderIndex(0)
-        else:
-            if self._active_extruder_index != -1:
-                self.setActiveExtruderIndex(-1)
-
-        self.activeExtruderChanged.emit()
+    def _globalContainerStackChanged(self) -> None:
+        # If the global container changed, the machine changed and might have extruders that were not registered yet
+        self._addCurrentMachineExtruders()
 
         self.resetSelectedObjectExtruders()
 
     ##  Adds the extruders of the currently active machine.
     def _addCurrentMachineExtruders(self) -> None:
-        global_stack = Application.getInstance().getGlobalContainerStack()
-        if global_stack and global_stack.getBottom():
-            self.addMachineExtruders(global_stack.getBottom(), global_stack.getId())
+        global_stack = self._application.getGlobalContainerStack()
+        extruders_changed = False
+
+        if global_stack:
+            container_registry = ContainerRegistry.getInstance()
+            global_stack_id = global_stack.getId()
+
+            # Gets the extruder trains that we just created as well as any that still existed.
+            extruder_trains = container_registry.findContainerStacks(type = "extruder_train", machine = global_stack_id)
+
+            # Make sure the extruder trains for the new machine can be placed in the set of sets
+            if global_stack_id not in self._extruder_trains:
+                self._extruder_trains[global_stack_id] = {}
+                extruders_changed = True
+
+            # Register the extruder trains by position
+            for extruder_train in extruder_trains:
+                self._extruder_trains[global_stack_id][extruder_train.getMetaDataEntry("position")] = extruder_train
+
+                # regardless of what the next stack is, we have to set it again, because of signal routing. ???
+                extruder_train.setParent(global_stack)
+                extruder_train.setNextStack(global_stack)
+                extruders_changed = True
+
+            self._fixMaterialDiameterAndNozzleSize(global_stack, extruder_trains)
+            if extruders_changed:
+                self.extrudersChanged.emit(global_stack_id)
+                self.setActiveExtruderIndex(0)
+
+    #
+    # This function tries to fix the problem with per-extruder-settable nozzle size and material diameter problems
+    # in early versions (3.0 - 3.2.1).
+    #
+    # In earlier versions, "nozzle size" and "material diameter" are only applicable to the complete machine, so all
+    # extruders share the same values. In this case, "nozzle size" and "material diameter" are saved in the
+    # GlobalStack's DefinitionChanges container.
+    #
+    # Later, we could have different "nozzle size" for each extruder, but "material diameter" could only be set for
+    # the entire machine. In this case, "nozzle size" should be saved in each ExtruderStack's DefinitionChanges, but
+    # "material diameter" still remains in the GlobalStack's DefinitionChanges.
+    #
+    # Lateer, both "nozzle size" and "material diameter" are settable per-extruder, and both settings should be saved
+    # in the ExtruderStack's DefinitionChanges.
+    #
+    # There were some bugs in upgrade so the data weren't saved correct as described above. This function tries fix
+    # this.
+    #
+    # One more thing is about material diameter and single-extrusion machines. Most single-extrusion machines don't
+    # specifically define their extruder definition, so they reuse "fdmextruder", but for those machines, they may
+    # define "material diameter = 1.75" in their machine definition, but in "fdmextruder", it's still "2.85". This
+    # causes a problem with incorrect default values.
+    #
+    # This is also fixed here in this way: If no "material diameter" is specified, it will look for the default value
+    # in both the Extruder's definition and the Global's definition. If 2 values don't match, we will use the value
+    # from the Global definition by setting it in the Extruder's DefinitionChanges container.
+    #
+    def _fixMaterialDiameterAndNozzleSize(self, global_stack, extruder_stack_list):
+        keys_to_copy = ["material_diameter", "machine_nozzle_size"]  # these will be copied over to all extruders
+
+        extruder_positions_to_update = set()
+        for extruder_stack in extruder_stack_list:
+            for key in keys_to_copy:
+                # Only copy the value when this extruder doesn't have the value.
+                if extruder_stack.definitionChanges.hasProperty(key, "value"):
+                    continue
+
+                setting_value_in_global_def_changes = global_stack.definitionChanges.getProperty(key, "value")
+                setting_value_in_global_def = global_stack.definition.getProperty(key, "value")
+                setting_value = setting_value_in_global_def
+                if setting_value_in_global_def_changes is not None:
+                    setting_value = setting_value_in_global_def_changes
+                if setting_value == extruder_stack.definition.getProperty(key, "value"):
+                    continue
+
+                setting_definition = global_stack.getSettingDefinition(key)
+                new_instance = SettingInstance(setting_definition, extruder_stack.definitionChanges)
+                new_instance.setProperty("value", setting_value)
+                new_instance.resetState()  # Ensure that the state is not seen as a user state.
+                extruder_stack.definitionChanges.addInstance(new_instance)
+                extruder_stack.definitionChanges.setDirty(True)
+
+                # Make sure the material diameter is up to date for the extruder stack.
+                if key == "material_diameter":
+                    position = int(extruder_stack.getMetaDataEntry("position"))
+                    extruder_positions_to_update.add(position)
+
+        # We have to remove those settings here because we know that those values have been copied to all
+        # the extruders at this point.
+        for key in keys_to_copy:
+            if global_stack.definitionChanges.hasProperty(key, "value"):
+                global_stack.definitionChanges.removeInstance(key, postpone_emit = True)
+
+        # Update material diameter for extruders
+        for position in extruder_positions_to_update:
+            self.updateMaterialForDiameter(position, global_stack = global_stack)
 
     ##  Get all extruder values for a certain setting.
     #
@@ -569,6 +495,8 @@ class ExtruderManager(QObject):
 
         result = []
         for extruder in ExtruderManager.getInstance().getMachineExtruders(global_stack.getId()):
+            if not extruder.isEnabled:
+                continue
             # only include values from extruders that are "active" for the current machine instance
             if int(extruder.getMetaDataEntry("position")) >= global_stack.getProperty("machine_extruder_count", "value"):
                 continue
@@ -632,13 +560,103 @@ class ExtruderManager(QObject):
     #
     #   This is exposed to qml for display purposes
     #
-    #   \param key The key of the setting to retieve values for.
+    #   \param key The key of the setting to retrieve values for.
     #
     #   \return String representing the extruder values
     @pyqtSlot(str, result="QVariant")
     def getInstanceExtruderValues(self, key):
         return ExtruderManager.getExtruderValues(key)
 
+    ##  Updates the material container to a material that matches the material diameter set for the printer
+    def updateMaterialForDiameter(self, extruder_position: int, global_stack = None):
+        if not global_stack:
+            global_stack = Application.getInstance().getGlobalContainerStack()
+            if not global_stack:
+                return
+
+        if not global_stack.getMetaDataEntry("has_materials", False):
+            return
+
+        extruder_stack = global_stack.extruders[str(extruder_position)]
+
+        material_diameter = extruder_stack.material.getProperty("material_diameter", "value")
+        if not material_diameter:
+            # in case of "empty" material
+            material_diameter = 0
+
+        material_approximate_diameter = str(round(material_diameter))
+        material_diameter = extruder_stack.definitionChanges.getProperty("material_diameter", "value")
+        setting_provider = extruder_stack
+        if not material_diameter:
+            if extruder_stack.definition.hasProperty("material_diameter", "value"):
+                material_diameter = extruder_stack.definition.getProperty("material_diameter", "value")
+            else:
+                material_diameter = global_stack.definition.getProperty("material_diameter", "value")
+                setting_provider = global_stack
+
+        if isinstance(material_diameter, SettingFunction):
+            material_diameter = material_diameter(setting_provider)
+
+        machine_approximate_diameter = str(round(material_diameter))
+
+        if material_approximate_diameter != machine_approximate_diameter:
+            Logger.log("i", "The the currently active material(s) do not match the diameter set for the printer. Finding alternatives.")
+
+            if global_stack.getMetaDataEntry("has_machine_materials", False):
+                materials_definition = global_stack.definition.getId()
+                has_material_variants = global_stack.getMetaDataEntry("has_variants", False)
+            else:
+                materials_definition = "fdmprinter"
+                has_material_variants = False
+
+            old_material = extruder_stack.material
+            search_criteria = {
+                "type": "material",
+                "approximate_diameter": machine_approximate_diameter,
+                "material": old_material.getMetaDataEntry("material", "value"),
+                "brand": old_material.getMetaDataEntry("brand", "value"),
+                "supplier": old_material.getMetaDataEntry("supplier", "value"),
+                "color_name": old_material.getMetaDataEntry("color_name", "value"),
+                "definition": materials_definition
+            }
+            if has_material_variants:
+                search_criteria["variant"] = extruder_stack.variant.getId()
+
+            container_registry = Application.getInstance().getContainerRegistry()
+            empty_material = container_registry.findInstanceContainers(id = "empty_material")[0]
+
+            if old_material == empty_material:
+                search_criteria.pop("material", None)
+                search_criteria.pop("supplier", None)
+                search_criteria.pop("brand", None)
+                search_criteria.pop("definition", None)
+                search_criteria["id"] = extruder_stack.getMetaDataEntry("preferred_material")
+
+            materials = container_registry.findInstanceContainers(**search_criteria)
+            if not materials:
+                # Same material with new diameter is not found, search for generic version of the same material type
+                search_criteria.pop("supplier", None)
+                search_criteria.pop("brand", None)
+                search_criteria["color_name"] = "Generic"
+                materials = container_registry.findInstanceContainers(**search_criteria)
+            if not materials:
+                # Generic material with new diameter is not found, search for preferred material
+                search_criteria.pop("color_name", None)
+                search_criteria.pop("material", None)
+                search_criteria["id"] = extruder_stack.getMetaDataEntry("preferred_material")
+                materials = container_registry.findInstanceContainers(**search_criteria)
+            if not materials:
+                # Preferred material with new diameter is not found, search for any material
+                search_criteria.pop("id", None)
+                materials = container_registry.findInstanceContainers(**search_criteria)
+            if not materials:
+                # Just use empty material as a final fallback
+                materials = [empty_material]
+
+            Logger.log("i", "Selecting new material: %s", materials[0].getId())
+
+            extruder_stack.material = materials[0]
+
     ##  Get the value for a setting from a specific extruder.
     #
     #   This is exposed to SettingFunction to use in value functions.
@@ -650,13 +668,16 @@ class ExtruderManager(QObject):
     #   global stack if not found.
     @staticmethod
     def getExtruderValue(extruder_index, key):
+        if extruder_index == -1:
+            extruder_index = int(Application.getInstance().getMachineManager().defaultExtruderPosition)
         extruder = ExtruderManager.getInstance().getExtruderStack(extruder_index)
 
         if extruder:
             value = extruder.getRawProperty(key, "value")
             if isinstance(value, SettingFunction):
                 value = value(extruder)
-        else: #Just a value from global.
+        else:
+            # Just a value from global.
             value = Application.getInstance().getGlobalContainerStack().getProperty(key, "value")
 
         return value
